 '***************************************************************
'*  Name    : SDFS2620_DEMO.BAS                                 *
'*  Author  : Original by MELabs Inc, Modified by Ranjit Diol   *
'*  Notice  : Copyright (c) 2011 COMPSys Workbench              *
'*          : All Rights Reserved                               *
'*  Date    : 10/15/2011                                        *
'*  Version : 1.1                                               *
'*  Notes   : This is modified version of the PIC BasicPro      *
'*          : sample code available at http://www.melabs.com    *
'****************************************************************
'                   NO WARRANTY IS IMPLIED                      *
'   DISCLAIMER: This file is being released as non-commericial
'   software. It is being provided "AS IS", neither the author, 
'   nor COMPSys Workbench shall be held liable for any damages or
'   hardships caused by its use.
' 
'----------------------------------------------------------------
' PICBASIC PRO 2.60L test program to work with MMC/SD cards using
' the SDMM3F or MMSD3F 3.3v development board.
' MCU: PIC18LF2620  (3.3v version)
' Remarks:
' Includes debug output @ 9600 baud for all I/O operations
' using the RS232 SEROUT2 statement. A zero indicates no error
' These can be removed or commented out as needed
'-----------------------------------------------------------------

	Define	OSC	40	' 40 MHz clock with HPLL config
'PIC 18LF2620 Configuration
'For HS mode:
'__CONFIG _CONFIG1H, _OSC_HS_1H &  _FCMEN_OFF_1H & _IESO_OFF_1H
'For HSPLL mode:
'__CONFIG _CONFIG1H, _OSC_HSPLL_1H &  _FCMEN_OFF_1H & _IESO_OFF_1H

#CONFIG
     __CONFIG _CONFIG1H, _OSC_HSPLL_1H &  _FCMEN_OFF_1H & _IESO_OFF_1H
     __CONFIG _CONFIG2H, _WDT_OFF_2H
     __CONFIG _CONFIG2L, _PWRT_OFF_2L & _BOREN_OFF_2L
     __CONFIG _CONFIG3H, _PBADEN_OFF_3H & _MCLRE_ON_3H
     __CONFIG _CONFIG4L, _LVP_OFF_4L & _XINST_OFF_4L
#ENDCONFIG

' Alias PIC pins and registers for SD/MMC card
' Some pins are not used in this version
'SD_WE          SD card write protect   *** NOT USED ***
'SD_WE_TRIS	    SD card write protect pin direction   *** NOT USED ***
'SDI		    SPI data in
'SDI_TRIS	    SPI data in direction
'SCL		    SPI clock
'SCL_TRIS	    SPI clock direction
'SD_CS		    SD card chip select
'SD_CS_TRIS	    SD card chip select direction
'SD_CD		    SD card detect   *** NOT USED ***
'SD_CD_TRIS	    SD card detect direction   *** NOT USED ***
'SDO		    SPI data out
'SDO_TRIS	    SPI data out direction

' Alias PIC pins and registers for SD/MMC card

'********************************************************************
'For SDMM3F and MMSD3F boards
'NOTE: we are not using the Write Enable pin on the card's socket.
 'Therefore, we will create a SD_WE variable and make FALSE (value=0)
'SD_WE		Var	PORTA.2	' SD card write protect
 SD_WE var byte
 SD_WE = 0 'FALSE
'Not used  
SD_WE_TRIS	Var	TRISA.2	' SD card write protect direction
'**********************************************************************

SDI		Var	PORTC.4	' SPI data in
SDI_TRIS	Var	TRISC.4	' SPI data in direction
SCL		Var	PORTC.3	' SPI clock
SCL_TRIS	Var	TRISC.3	' SPI clock direction
SD_CS		Var	PORTC.2	' SD card chip select
SD_CS_TRIS	Var	TRISC.2	' SD card chip select direction
SD_CD		Var	PORTA.1	' SD card detect
SD_CD_TRIS	Var	TRISA.1	' SD card detect direction
SDO		Var	PORTC.5	' SPI data out
SDO_TRIS	Var	TRISC.5	' SPI data out direction

x var word
ch var byte


' Include the SD/MMC subroutines
	Include "SDFS.BAS"
	SDC_UseHardSPI = TRUE	' Use hardware SSP port for SPI.
    
    'We force SD_WE to be FALSE ...seee note above
    SD_WE = FALSE
    
	ADCON1 = 15		' All I/O pins digital
	Pause 100
'+++++++++++++++++++++++++++++++++++++++++++++++++
'Start of main code
    Serout2 PORTC.6, 84, ["****** SDFS demo code Ver 1.1 *******",$d,$a]
	Serout2 PORTC.6, 84, ["SD File System Demo Code using",$d,$a,"the MMSD3 or SDMM3 3.3v board",$d,$a]
	Serout2 PORTC.6, 84, ["Original sample code available at http://www.melabs.com",$d,$a]	
    Serout2 PORTC.6, 84, ["*************************************",$d,$a]
' FSInit initializes the card and reads all the preliminary information from it
	Gosub FSInit
	
 'Debug: Output error code	
	Serout2 PORTC.6, 84, [$d,$a,"FSInit: ", Dec FAT_error, " ", Dec SDC_status, " ", Dec SDC_response, $d, $a]

	If (FAT_error != 0) Then Stop

	' Display card directory
	Serout2 PORTC.6, 84, ["Directory of files:",$d,$a]
	Gosub FINDfirst		' Find first file on card
	While (FAT_error = 0)
		Serout2 PORTC.6, 84, [Str FAT_FileName\11, $d, $a]
		Gosub FINDnext	' Find next file on card
	Wend

' This section defines a specific short (8.3) filename
'  Note that spaces are use in empty elements and must be upper case for Windows
	'FAT_FileName[0] = "T"
	'FAT_FileName[1] = "E"
	'FAT_FileName[2] = "S"
	'FAT_FileName[3] = "T"
	'FAT_FileName[4] = "0"
	'FAT_FileName[5] = "1"
	'FAT_FileName[6] = " "
	'FAT_FileName[7] = " "
	'FAT_FileName[8] = "T"
	'FAT_FileName[9] = "X"
	'FAT_FileName[10] = "T"

'Alternative method
'Load file name using a FOR/NEXT loop 

'Load file name    
    for x = 0 to 10
        lookup x,["TEST01  TXT"],ch   'Note: must be 11 chars
        FAT_FileName[x] = ch
    next x 	

' Set file time to 8:30:10 and date to 10/12/2011
' Year (0 = 1980, 127 = 2107)
' Seconds are in 2 second resolution Ex: 5 = 10 seconds (5x2)

	FAT_seconds = 5
	FAT_minutes = 30
	FAT_hours = 8
	FAT_day = 12
	FAT_month = 10
	FAT_year = 31

' Open a file for write
	FAT_mode = "w"		' Write mode
	Gosub FSfopen		' Open file pointed to by Byte array FAT_FileName

 'Debug: Output error code
	Serout2 PORTC.6, 84, ["Open for write: ", Dec FAT_error, $d, $a]
	If (FAT_error != 0) Then Stop

' Write data to the file  FAT_src[x] is a 512 byte buffer
    for x = 0 to 36
        lookup x,["Hello World!",$d,$a,"Welcome to programming!"],ch
        FAT_src[x] = ch
    next x 
    FAT_count = x 'in this case x=12 which is the number of characters to be written
	Gosub FSfwrite
 
 'Debug: Output error code
	Serout2 PORTC.6, 84, [ "FSWrite ", Dec FAT_error, $d, $a]
	If (FAT_error != 0) Then Stop

' Close file
	Gosub FSfclose
	
 'Debug: Output error code	
	Serout2 PORTC.6, 84, [ "FSClose ", Dec FAT_error, $d, $a]
	If (FAT_error != 0) Then Stop

' Open a file for read
	FAT_mode = "r"		' Read mode
	Gosub FSfopen		' Open file pointed to by Byte array FAT_FileName

 'Debug: Output error code
	Serout2 PORTC.6, 84, ["FSOpen: ", Dec FAT_error, $d, $a]
	If (FAT_error != 0) Then Stop

' Read and display the whole file
	FAT_count = 1	' Read 1 byte to buffer at a time
	Gosub FSfread
	While (FAT_error = 0)
		Serout2 PORTC.6, 84, [FAT_dest[0]]
		FAT_count = 1	' Read 1 byte to buffer at a time
		Gosub FSfread
	Wend

 'Debug: Output error code	
	Serout2 PORTC.6, 84, [ $d, $a, "FSRead: ", Dec FAT_error]
    if FAT_error = CE_EOF then
       Serout2 PORTC.6, 84, [" (End Of File)",$d,$a]
    endif   

'**************************************************************************
' The following code is commented out for the purpose of this demo.
' You may use the following routines to RENAME and DELETE files.
'
' To Delete a file
'	Gosub FSremove		' Delete file pointed to by Byte array FAT_FileName
'	Serout2 PORTC.6, 84, ["Remove: ", Dec FAT_error, $d, $a]
'	If (FAT_error != 0) Then Stop

' Rename a file
' FAT_FileName2 is new name
'	FAT_FileName2[0] = "T"
'	FAT_FileName2[1] = "E"
'	FAT_FileName2[2] = "S"
'	FAT_FileName2[3] = "T"
'	FAT_FileName2[4] = "2"
'	FAT_FileName2[5] = " "
'	FAT_FileName2[6] = " "
'	FAT_FileName2[7] = " "
'	FAT_FileName2[8] = "T"
'	FAT_FileName2[9] = "X"
'	FAT_FileName2[10] = "T"
'
'Alternative method
'Load file name using a FOR/NEXT loop   
'    for x = 0 to 10
'       lookup x,["TEST2   TXT"],ch   'Note: must be 11 chars
'        FAT_FileName[x] = ch
'   next x 	
'	Gosub FSrename		' Rename file pointed to by Byte array FAT_FileName
'	Serout2 PORTC.6, 84, ["Rename: ", Dec FAT_error, $d, $a]
'	If (FAT_error != 0) Then Stop

	End
